/*
    Copyright (C) 1996-1998  Ulric Eriksson <ulric@edu.stockholm.se>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <sys/wait.h>
#include "cmalloc.h"
#include "common.h"

/* ---
common.c

Device independent functions for all three programs.
--- */

/* ---
Strip trailing newline, if there is one
*/

void chomp(char *p)
{
	if ((p = strchr(p, '\n'))) *p = '\0';
}

/* ---
Check if a string should be interpreted as a positive answer.
*/

int itsayes(char *p)
{
	return !cstrcasecmp(p, "yes")
		|| !cstrcasecmp(p, "1")
		|| !cstrcasecmp(p, "on")
		|| !cstrcasecmp(p, translate("yes"))
		|| !cstrcasecmp(p, translate("on"));
}

/* ---
Case insensitive compare
*/

int cstrcasecmp(const char *p, const char *q)
{
	int c;

	while (!(c = toupper(*p)-toupper(*q)) && *p) {
		p++;
		q++;
	}
	return c;
}

/* ---
Case insensitive compare, length limited
*/

int cstrncasecmp(const char *p, const char *q, size_t n)
{
	size_t i = 0;
	int c = 0;

	while ((i < n) && !(c = toupper(*p)-toupper(*q)) && *p) {
		p++;
		q++;
		i++;
	}
	return c;
}

/* ---
Delete all temp directories whose processes are no longer around.
NB: Doesn't do anything at all.
*/

void stale_dirs(char *dn)
{
}

/* a little bookkeeping to keep track of temp files and make sure they
   are deleted when child processes terminate */

static int ndeletia;
static struct {
	long p;
	char *fn;
} *deletia;

/* ---
associate a temp file with a process
*/

void deletia_add(long p, char *fn)
{
	int i;

	for (i = 0; i < ndeletia; i++)
		if (deletia[i].fn == NULL) break;
	if (i == ndeletia) {
		ndeletia++;
		deletia = crealloc(deletia, ndeletia*sizeof *deletia);
	}
	deletia[i].p = p;
	deletia[i].fn = cstrdup(fn);
}

/* ---
mark all files related to a process to be deleted later
p == 0 marks all files
*/

void deletia_mark(long p)
{
	int i;

	for (i = 0; i < ndeletia; i++) {
		if (p == 0 || p == deletia[i].p)
			deletia[i].p = 0;
	}
}

/* ---
remove files that have previously been marked
*/

void deletia_reap(void)
{
	int i;

	for (i = 0; i < ndeletia; i++) {
		if (deletia[i].p == 0 && deletia[i].fn) {
			remove(deletia[i].fn);
			cfree(deletia[i].fn);
			deletia[i].fn = NULL;
		}
	}
}

/* ---
Signal handler for dead children.
*/

void waitforchild(int i)
{
	pid_t p;
	long lp;

	p = waitpid(-1, NULL, WNOHANG);
	lp = (long)p;
	if (p > 0) deletia_mark(lp);
	signal(SIGCHLD, waitforchild);
}

static struct {
	char *key, *xl;
} *dict = NULL;

static long nw = 0;

/* ---
Load the message string dictionary from $(SIAGHOME)/common/dictionary.$(LANG)
*/

void load_dictionary(char *fn)
{
        FILE *fp = fopen(fn, "r");
        char b[1024], *p;

        if (!fp) {
                fprintf(stderr, "Can't open dictionary\n");
                return;
        }
        while (fgets(b, sizeof b, fp)) {
                chomp(b);
                p = strchr(b, '\t');
                if (!p) continue;
                *p++ = 0;
                p += strspn(p, "\t");
                dict = crealloc(dict, (nw+1)*(sizeof dict[0]));
                dict[nw].key = cstrdup(b);
                dict[nw++].xl = cstrdup(p);
        }
}

/* ---
A function that doesn't do anything. Can be used to set break points
during debugging.
*/

void dummy_fun(void)
{
	;
}

/* ---
Translate a message string. If no match is found, the key is returned.
By using English keys, an application can keep working even when the
dictionary is incomplete.
*/

char *translate(char *key)
{
        long i;

        for (i = 0; i < nw; i++)
                if (!strcmp(key, dict[i].key)) {
			return dict[i].xl;
		}
        return key;
}


