/*
   Pathetic Writer
   Copyright (C) 1997, 1998  Ulric Eriksson <ulric@edu.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
 */

/*
 * matrix.c
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "../common/common.h"
#include "../common/cmalloc.h"

#include "pw.h"

int max_lines = BUFFER_ROWS;
int max_columns = BUFFER_COLS;


/* ---
Style code for PW. The styles are quite different from the ones in Siag,
but the principles are the same: Styles are encoded to integers so that
they can be handled by the general format code.
--- */

int nstyle;
style *style_table = NULL;

/* ---
*/
char *style2name(int i)
{
	if (i < 0 || i >= nstyle) return NULL;
	return style_table[i].name;
}

/* ---
*/
int name2style(char *name)
{
	int i;

	if (!name) return -1;
	for (i = 0; i < nstyle; i++) {
		if (!cstrcasecmp(name, style_table[i].name)) return i;
	}
	return -1;
}

static struct bp_style {
	sfmt fmt;	/* the format we haven't initialized */
	int n;		/* index into style_table */
} *bp_styles = NULL;
static int nbp_styles = 0;	/* size of backlog */

static void add_bp_style(int n, sfmt fmt)
{
	int i = nbp_styles++;
	bp_styles = crealloc(bp_styles, nbp_styles*sizeof *bp_styles);
	bp_styles[i].fmt = fmt;
	bp_styles[i].fmt.family = cstrdup(fmt.family);
	bp_styles[i].fmt.fg = cstrdup(fmt.fg);
	bp_styles[i].fmt.bg = cstrdup(fmt.bg);
	bp_styles[i].n = n;
}

/* ---
*/
void do_bp_styles(void)
{
	int i;

	for (i = 0; i < nbp_styles; i++) {
		style_table[bp_styles[i].n].format =
			encode_format(~0, &bp_styles[i].fmt);
		cfree(bp_styles[i].fmt.family);
		cfree(bp_styles[i].fmt.fg);
		cfree(bp_styles[i].fmt.bg);
	}
	cfree(bp_styles);
}

/* ---
A struct style has the fields name, format index, format, next.
The format has the fields family, size, bold, italic, uline, fg, bg, style.

I omit the format index because we will compute it here.
The style in the format is useless here.
Finally, we don't use the bg colour.

The function still takes an awful lot of arguments.

A dilemma: this function is called from register-style before the ui
has been initialised. But we need to look up colours and fonts, which
won't work until we have a valid Display.

Why not run this *after* init_windows? Because documents are also
loaded before init_windows, so we must have valid styles. Besides,
load_pw also calls lookup_style.

My solution: when ok2print is false, don't look up fonts and colours
but save them in a separate array. Then look them up by calling
the function patch_styles from init_windows.
*/

int lookup_style(char *name, char *next,
		char *family, int size, int bold, int italic,
		int uline, char *fg)
{
	sfmt fmt;
	style s;
	int i;

	s.name = cstrdup(name);

	/* get the default format, so all fields are valid */
	decode_format(0, ~0, &fmt);
	/* then change the ones we are interested in */
	fmt.family = family;
	fmt.size = size;
	fmt.bold = bold;
	fmt.italic = italic;
	fmt.uline = uline;
	fmt.fg = fg;

	/* now figure out where to put it */
	i = name2style(name);
	if (i == -1) {	/* new style */
		i = nstyle++;
		style_table = crealloc(style_table, nstyle*sizeof *style_table);
	} else {	/* redefine existing */
		cfree(style_table[i].name);
	}

	if (ok2print) {		/* UI is initialised */
		s.format = encode_format(~0, &fmt);
	} else {		/* do it later, then */
		s.format = 0;
		add_bp_style(i, fmt);
	}

	style_table[i] = s;
	if ((style_table[i].follower = name2style(next)) == -1)
		style_table[i].follower = 0;	/* always valid */
	return i;
}

static int compar(const void *p, const void *q)
{
	return strcmp(*(const char **)p, *(const char **)q);
}

/* ---
*/
char **style_list(int *n)
{
	int i;
	char **list = (char **)cmalloc((nstyle+1)*sizeof(char *));
	for (i = 0; i < nstyle; i++)
		list[i] = cstrdup(style_table[i].name);
	list[i] = NULL;
	qsort(list, i, sizeof(char *), compar);
	*n = i;
	return list;
}

/* ---
*/
int style_height(int s)
{
	int fmt = style_table[s].format;
	int font = format_table[fmt].font;
	return ps_font_height(font);
}

/* ---
this should work exactly like the format code in fonts.c, because I don't
know what fields will be added or removed in the future.

the styles are saved in a single level, just like the formats.
Attributes: name, height, next, family, size, bold, italic, uline, fg
*/

void save_styles(FILE *fp, int i)
{
	sfmt fmt;
	style st;

	st = style_table[i];
	decode_format(st.format, ~0, &fmt);
	fprintf(fp, ".style %d\n", i);
	fprintf(fp, "name %s\n", st.name);
	fprintf(fp, "next %d\n", st.follower);
	fprintf(fp, "family %s\n", fmt.family);
	fprintf(fp, "size %d\n", fmt.size);
	fprintf(fp, "bold %d\n", fmt.bold);
	fprintf(fp, "italic %d\n", fmt.italic);
	fprintf(fp, "uline %d\n", fmt.uline);
	fprintf(fp, "fg %s\n", fmt.fg);
	fprintf(fp, "end\n");
}

/* ---
*/
int load_styles(FILE *fp)
{
	style st = style_table[0];
	sfmt fmt;

	char name[1000], family[1000], fg[1000], b[1000];
	decode_format(0, ~0, &fmt);
	strcpy(name, st.name);
	st.name = name;
	strcpy(family, fmt.family);
	fmt.family = family;
	strcpy(fg, fmt.fg);
	fmt.fg = fg;
	while (fgets(b, sizeof b, fp) && strncmp(b, "end", 3)) {
		chomp(b);
		if (!strncmp(b, "name ", 5))
			strcpy(st.name, b+5);
		else if (!strncmp(b, "next ", 5))
			st.follower = name2style(b+5);
		else if (!strncmp(b, "family ", 7))
			strcpy(fmt.family, b+7);
		else if (!strncmp(b, "size ", 5))
			fmt.size = strtol(b+5, NULL, 10);
		else if (!strncmp(b, "bold ", 5))
			fmt.bold = strtol(b+5, NULL, 10);
		else if (!strncmp(b, "italic ", 7))
			fmt.italic = strtol(b+7, NULL, 10);
		else if (!strncmp(b, "uline ", 6))
			fmt.uline = strtol(b+6, NULL, 10);
		else if (!strncmp(b, "fg ", 3))
			strcpy(fmt.fg, b+3);
		/* ignore anything we don't grok */
	}
	return lookup_style(st.name, style2name(st.follower),
			fmt.family, fmt.size, fmt.bold, fmt.italic,
			fmt.uline, fmt.fg);
}

/* end of style code */


/* ---
   Frees the memory used by the matrix.
*/

void free_text(rich_text *c)
{
	cfree(c);
}

static rich_char empty_char = {
	'\0', 0
};

#ifndef MAX
#define MAX(a,b) ((a)>(b)?(a):(b))
#endif


/* ---
make sure the char is allocated
note that this function changes the buffer
If row is greater than any previously used line, new lines are allocated
up to and including row. The new lines are initialized to be empty.
*/

void alloc_line(buffer *b, int sht, long row)
{
	long i = b->sht[sht].used_lines;
	if (row <= b->sht[sht].alloc_lines) {
		return;
	}
	b->sht[sht].alloc_lines = b->sht[sht].used_lines = MAX(b->sht[sht].used_lines, row);
	b->sht[sht].text = (rich_text *)crealloc(b->sht[sht].text,
			(b->sht[sht].alloc_lines+1)*sizeof(rich_text));
	i++;	/* don't overwrite the last used line */
	while (i <= b->sht[sht].alloc_lines) {
		b->sht[sht].text[i].height = style_height(0);
		b->sht[sht].text[i].sty = 0;
		b->sht[sht].text[i].adj = HADJ_LEFT;
		b->sht[sht].text[i++].p = NULL;
	}
	b->sht[sht].used_lines = b->sht[sht].alloc_lines;
}

/* ---
kludge to keep things from breaking fatally
any positions after row r are reset to (r,0)
NB: this function does not take care of point
*/

void position_kludge(buffer *b, int sht, int r)
{
	window *w = w_list;
	position p = make_position(r, 0);

	do {
		if (w->buf == b) {
			sheet *s = w->buf->sht;
			if (s[sht].top.row >= r) s[sht].top = p;
			if (s[sht].blku.row >= r) s[sht].blku = p;
			if (s[sht].blkl.row >= r) s[sht].blkl = p;
		}
		w = w->next;
	} while (w != w_list);

	if (b->sht[sht].mark_pos.row >= r) b->sht[sht].mark_pos = p;
}

#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif

/* ---
this one only resets positions that don't exist
it does take care of everything, including point
*/

void position_kludge2(void)
{
	window *w = w_list;
	buffer *b;
	unsigned long r;
	int s;
	sheet *st;

	do {
		s = w->sht;
		b = w->buf;
		st = b->sht;
		r = line_last_used(b, s);
		if (r < 1) r = 1;
		st[s].top.row = MIN(st[s].top.row, r);
		st[s].top.col = MIN(st[s].top.col,
				col_last_used(b, s, st[s].top.row));
		st[s].blku.row = MIN(st[s].blku.row, r);
		st[s].blku.col = MIN(st[s].blku.col,
				col_last_used(b, s, st[s].blku.row));
		st[s].blkl.row = MIN(st[s].blkl.row, r);
		st[s].blkl.col = MIN(st[s].blkl.col,
				col_last_used(b, s, st[s].blkl.row));
		st[s].point_pos.row = MIN(st[s].point_pos.row, r);
		st[s].point_pos.col = MIN(st[s].point_pos.col,
				col_last_used(b, s, st[s].point_pos.row));
		w = w->next;
	} while (w != w_list);
	b = b_list;
	do {
		for (s = 0; s < b->nsht; s++) {
			r = line_last_used(b, s);
			if (r < 1) r = 1;
			b->sht[s].mark_pos.row = MIN(b->sht[s].mark_pos.row, r);
			b->sht[s].mark_pos.col = MIN(b->sht[s].mark_pos.col,
					col_last_used(b, s, b->sht[s].mark_pos.row));
		}
		b = b->next;
	} while (b != b_list);
}

/* ---
shift the next line down, make it the follower's style,
copy the remainder of the line and truncate this line
*/

int split_line(buffer *buf, int s, long row, long col)
{
	position_kludge(buf, s, row);

	alloc_line(buf, s, row);
	if (buf->sht[s].text[row].sty == STY_EMBED) {
		if (col == 0) downshift_text(buf, s, row);
		else downshift_text(buf, s, row+1);
		return 1;
	}

	downshift_text(buf, s, row+1);	/* make room for the new line */
	buf->sht[s].text[row+1].sty = style_table[buf->sht[s].text[row].sty].follower;
	buf->sht[s].text[row+1].p = rc_strdup(buf->sht[s].text[row].p+col);
	if (rc_strlen(buf->sht[s].text[row].p) > col)
		buf->sht[s].text[row].p[col] = empty_char;	/* truncate */
	/* check line heights for both lines */
	check_line_height(buf, s, row);
	check_line_height(buf, s, row+1);
	buf->change = TRUE;
	return 1;
}

/* ---
*/
int join_lines(buffer *buf, int s, long row)
{
	rich_char *p;
	position_kludge(buf, s, row);

	if (row >= buf->sht[s].used_lines) return 0;	/* nothing to join */

	/* special case if either line is embedded object */
	if (buf->sht[s].text[row].sty == STY_EMBED) {
		/* can only join with empty line */
		if (line_length(buf, s, row+1) == 0) {
			upshift_text(buf, s, row+1);
			return 1;
		} else return 0;	/* can't do that */
	}

	if (buf->sht[s].text[row+1].sty == STY_EMBED) {
		if (line_length(buf, s, row) == 0) {
			upshift_text(buf, s, row);
			return 1;
		} else return 0;
	}

	p = buf->sht[s].text[row].p;
	buf->sht[s].text[row].p = rc_strins(p, buf->sht[s].text[row+1].p, rc_strlen(p));
	cfree(p);
	upshift_text(buf, s, row+1);
	return 1;
}

/* ---
*/
int ins_text(buffer *buf, int s, position pos, unsigned char *t, long fmt)
{
	rich_char *p;
	rich_char *q;
	if (ret_style(buf, s, pos.row) == STY_EMBED) return 0;	/* nope */

	alloc_line(buf, s, pos.row);
	p = buf->sht[s].text[pos.row].p;
	q = rc_makerich(t, fmt);
	buf->sht[s].text[pos.row].p = rc_strins(p, q, pos.col);
	check_line_height(buf, s, pos.row);
	cfree(p);
	cfree(q);
	return 1;
}

/* ---
*/
int ins_char(buffer *buf, int sht, long r, long c, int s, long fmt)
{
	unsigned char b[2];
	b[0] = s;
	b[1] = '\0';
	return ins_text(buf, sht, make_position(r, c), b, fmt);
}

/* ---
Make a plaintext copy of row r int the string p.
The string must be freed by caller
*/

unsigned char *peek_line(buffer *b, int s, long r)
{
	if (b == NULL) return NULL;
	if (r > b->sht[s].used_lines) return NULL;
	if (b->sht[s].text == NULL) return NULL;
	return rc_makeplain(b->sht[s].text[r].p);
}

/* ---
*/
int peek_char(buffer *b, int s, long r, long c)
{
	int result;
	unsigned char *p = peek_line(b, s, r);
	result = p[c];
	cfree(p);
	return result;
}

/* ---
Figure out where to break a line
*/

static int find_suitable_space(buffer *b, int s, long r)
{
	long c = 0;
	long space = 0;	/* current "suitable" space */

	while (c < line_length(b, s, r) &&
			line_width(b, s, r, c) <=
				paper_width-left_margin-right_margin) {
		if (isspace(peek_char(b, s, r, c))) space = c;
		c++;
	}
	if (!space) {
#ifdef DEBUG
fprintf(stderr, "Warning: very long line (%ld chars) without spaces\n", c);
#endif
	}
	return space;
}

/* ---
Checks if the line is too long for the paper width and breaks it if
necessary. Continues with the next line.
*/

long rebreak_line(buffer *buf, int s, long row)
{
	while (line_width(buf, s, row, line_length(buf, s, row)) >
			paper_width-left_margin-right_margin) {
		int col = find_suitable_space(buf, s, row);
		if (col <= 0) break;	/* none found */
		split_line(buf, s, row, col);
		row++;
		del_char(buf, s, row, 0);
		/* We now have a fragment from the last part of the newly
		   broken line. If the next line is of the same style and
		   also contains a non-empty fragment, we will want to
		   join the lines and potentially rebreak the result */
		if ((buf->sht[s].text[row].sty == buf->sht[s].text[row+1].sty) &&
				(line_length(buf, s, row+1) > 0)) {
			ins_char(buf, s, row+1, 0, ' ', ret_format(buf, s, row, 0));
			join_lines(buf, s, row);
		}
		pr_scr_flag = TRUE;
	}
	return row;
}

extern void embed_unload(char *);	/* FIXME */

/* ---
returns number of deleted characters
*/

int del_char(buffer *buf, int s, long row, long col)
{
	if (buf->sht[s].text[row].sty == STY_EMBED) {
		if (col == 0) {		/* remove the object and the line */
			char *p = (char *)rc_makeplain(buf->sht[s].text[row].p);
			embed_unload(p);	/* can't be NULL */
			cfree(p);
			upshift_text(buf, s, row);
			return 1;
		} else return 0;
	}

	rc_strcpy(buf->sht[s].text[row].p+col, buf->sht[s].text[row].p+col+1);
	buf->change = TRUE;
	return 1;
}

/* ---
returns number of deleted characters
*/

long del_text(buffer *buf, int s, position p, long length)
{
	long i;

	for (i = 0; i < length; i++)
		if (!del_char(buf, s, p.row, p.col)) break;
	return i;
}

/* ---
*/
long del_lines(buffer *buf, int s, long row, long count)
{
	long i;

	/* don't delete lines that aren't there */
	if (row+count-1 > buf->sht[s].alloc_lines) count = buf->sht[s].alloc_lines-row+1;

	/* delete count lines, starting at row */
	for (i = 0; i < count; i++) {
		if (buf->sht[s].text[row+i].sty == STY_EMBED) {
			char *p = (char *)rc_makeplain(buf->sht[s].text[row].p);
			embed_unload(p);
			cfree(p);
		}
		cfree(buf->sht[s].text[row+i].p);
	}

	/* shift the rest of the lines up */
	for (i = row; i+count <= buf->sht[s].alloc_lines; i++) {
		buf->sht[s].text[i] = buf->sht[s].text[i+count];
	}

	/* clear the last count lines */
	for (; i <= buf->sht[s].alloc_lines; i++) {
		buf->sht[s].text[i].p = NULL;
	}
	buf->sht[s].alloc_lines -= count;
	buf->sht[s].used_lines -= count;
	if (count) buf->change = TRUE;
	return count;
}

/* ---
*/
int ins_format(buffer *buf, int s, long row, long c1, long c2, long format)
{
	rich_char *line;
	alloc_line(buf, s, row);
	line = buf->sht[s].text[row].p;
	while (c1 < c2 && c1 < rc_strlen(line)) {
		line[c1++].fmt = format;
	}
	return 1;
}

/* ---
*/
int ret_format(buffer *buf, int s, long row, long col)
{
	rich_char *line;
	if (row > buf->sht[s].used_lines) return style_table[0].format;
	line = buf->sht[s].text[row].p;
	if (!line || line[0].c == '\0') {
		/* empty line, try to get from previous line(s) */
		int pr = row-1;
		while (pr >= 1 && ret_style(buf, s, pr) == ret_style(buf, s, row)) {
			line = buf->sht[s].text[pr].p;
			col = rc_strlen(line);
			if (col) return line[col-1].fmt;
			pr--;
		}
		/* give up and return the default format for the style */
		return style_table[buf->sht[s].text[row].sty].format;
	}
	if (col >= rc_strlen(line)) col = rc_strlen(line)-1;
	return line[col].fmt;
}

/* ---
*/
void set_style(buffer *b, int s, long row, int sty)
{
	alloc_line(b, s, row);
	b->sht[s].text[row].sty = sty;
	b->sht[s].text[row].height = style_height(sty);
}

/* ---
*/
int ret_style(buffer *b, int s, long row)
{
	if (row > b->sht[s].used_lines) return STY_DEFAULT;
	return b->sht[s].text[row].sty;
}

/* ---
*/
int ret_hadj(buffer *b, int s, long row)
{
	if (row > b->sht[s].used_lines) return HADJ_LEFT;
	return b->sht[s].text[row].adj & HADJ_MASK;
}

/* ---
*/
long line_last_used(buffer *buf, int s)
{
	return buf->sht[s].used_lines;
}

/* ---
this means strlen
*/

long col_last_used(buffer *buf, int s, long row)
{
	return line_length(buf, s, row);	/* hmm */
}

/* ---
*/
void downshift_text(buffer *b, int s, long row)
{
	long r;
	int i;
	b->sht[s].used_lines++;
	alloc_line(b, s, b->sht[s].used_lines);
	for (r = b->sht[s].used_lines; r > row; r--) {
		b->sht[s].text[r] = b->sht[s].text[r-1];
	}
	b->sht[s].text[r].sty = 0;
	b->sht[s].text[r].height = style_height(0);
	b->sht[s].text[r].adj = HADJ_LEFT;
	b->sht[s].text[r].p = NULL;
	for (i = 0; i < b->sht[s].nplugin; i++) {
		if (b->sht[s].plugin[i].row >= row) b->sht[s].plugin[i].row++;
	}
	b->change = pr_scr_flag = TRUE;
}

/* ---
*/
void upshift_text(buffer *b, int s, long row)
{
	long r;
	int i, j;
	b->sht[s].used_lines--;
#ifdef	WANNADIE
This causes problems with other stuff like join_lines that wants to keep
the old data from b->pwdoc[row]

	if (b->pwdoc[row])
		cfree_pwline(b->pwdoc[row]);
#endif
	for (r = row; r <= b->sht[s].used_lines; r++) {
		b->sht[s].text[r] = b->sht[s].text[r+1];
	}
	/* this line is still allocated. We clear it just in case */
	b->sht[s].text[r].sty = 0;
        b->sht[s].text[r].height = style_height(0);
        b->sht[s].text[r].adj = HADJ_LEFT;
	b->sht[s].text[r].p = NULL;
	for (i = 0; i < b->sht[s].nplugin; i++) {
		if (b->sht[s].plugin[i].row == row) {
			pw_plugin_stop(b->sht[s].plugin[i].ph);
			cfree(b->sht[s].plugin[i].name);
			b->sht[s].nplugin--;
			for (j = i; j < b->sht[s].nplugin; j++)
				b->sht[s].plugin[j] = b->sht[s].plugin[j+1];
		} else if (b->sht[s].plugin[i].row > row) {
			b->sht[s].plugin[i].row--;
		}
	}
	b->change = pr_scr_flag = TRUE;
}

